/*
Copyright 2023 Alexandre Trofimov

This file is part of getcdmach external ballistics software.

getcdmach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

getcdmach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with getcdmach.  If not, see <http://www.gnu.org/licenses/>.
*/

const RMT_FULL = // reference Mach table = point of data in resulting Cd / Mach array
[0,
0.4,
0.5,
0.6,
0.7,
0.8,
0.825,
0.85,
0.875,
0.9,
0.925,
0.95,
0.975,
1,
1.025,
1.05,
1.075,
1.1,
1.15,
1.2,
1.3,
1.4,
1.5,
1.6,
1.8,
2,
2.2,
2.4,
2.6,
3,
4,
5];

// special version of ebalengine's Bullet object, for blanket same drag for any Mach
// implements same methods
function NeeBullet (cal, mass, cd) {
    this.cal = cal / 1000;
    this.mass = mass / 1000;
    this.cd = cd;
    this.bc = (this.mass / this.cal / this.cal) / (0.4535924 / 0.0254 / 0.0254);
    this.getDragForMach = function(mach) {
            return this.cd / this.bc * 2.08551e-04;
        } // getDragForMach method
}

// special version of ebalengine's Atmosphere object, for blanket std ICAO alt.0
// implements same methods
function NeeAtmo() {
    var stdatmo = new Atmosphere();
    stdatmo.setStd(0);
    var dss = stdatmo.getDensityAndSpeedOfSound(0);
    this.speedOfSound = dss.speedOfSound;
    this.density = dss.density;

    // this.ATMOS_DENSSTD = 0.076474;
    this.ATMOS_DENSSTD = 0.0764721232677921;

    this.getSpeedOfSound = function(altDelta) {
        return this.speedOfSound;
    }

    this.getDensity = function(altDelta) {
        return this.density;
    }

    this.getDensityAndSpeedOfSound = function(altDelta) {
        return { "density": this.density, "speedOfSound": this.speedOfSound };
    }
}

// special version of ebalengine's MuzzleVelocity object, for same constant mv, regardless of atmo, temperature or distance to chrono
// implements same methods
function NeeMuzzleVelocity(cv0) {
    this.v0 = cv0; // fps

    this.getRealV0 = function(catmo, bullet) {
            return this.v0;
        }
}


function VelTabPt(distance, velocity) {
    this.d = distance;
    this.v = velocity;
}

function CdMachPt(cd, mach) {
    this.cd = cd;
    this.mach = mach;
}

function CdMachTable(vtdata, cal, mass) {
    this.vtdata = vtdata;

    // calculate Cd/Mach table
    var cdmach = [];

    var bullet = new NeeBullet(cal, mass, 0);
    var atmo = new NeeAtmo();
    var rifle = new Rifle(0, 0);
    var shotin = new ShotInitials(0, 0, 0, 0); // no cant, no elevation, no azimuth delta
    var windo = new Wind(0, 0);
    var mv; // muzzle velocity
    var dd = 0; // distance delta
    var dasolver; // ballistic solver object
    var traj; // resulting trajectory
    var dv; // difference between expected and calculated velocity at the end of segment
    var vavg; // average speed over the segment
    var mach;
    
    var CD_PRECISION = 0.0001; // as fraction of the running cd absolute value
    var ITERCOUNT_MAX = 100; // sanity check; it should converge with < 30
    
    for (let i = 0; i < this.vtdata.length - 1; i = i + 1) {
        mv = new NeeMuzzleVelocity(this.vtdata[i].v);
        dd = this.vtdata[i+1].d - this.vtdata[i].d;
        dasolver = new Solver(rifle, bullet, mv);

        var cd = 0.01; // to start with
        var cdstep = 0.5;
        var itercount = 1;

        do { // cd iterations loop
            cd = cd + cdstep;
            dasolver.bullet.cd = cd;

            traj = dasolver.calculateTrajectory(
                atmo,
                shotin,
                windo,
                dd,
                dd
            );

            dv = this.vtdata[i+1].v - traj[1].velocity;

            // console.log("dv = " + dv + " - cd = " + cd + " - step = " + cdstep);

            if (Math.sign(dv) == Math.sign(cdstep)) {
                if (Math.abs(cdstep / cd) < CD_PRECISION) {
                    vavg = traj[1].range / traj[1].time;
                    break;
                }
                cdstep = -cdstep / 2;
            }

            itercount = itercount + 1;
            if (itercount > ITERCOUNT_MAX) { // sanity check
                throw ("Exceeded max number of iterations (" + ITERCOUNT_MAX + "). Bombing out.");
            }
        } while (true); // cd iterations loop, until we break out with needed cd precision reached

        mach = vavg / atmo.speedOfSound;

        console.log("CD at Mach " + mach + " (" + vavg + "fps) converged to " + cd + " in " + itercount + " iterations; " + "dv = " + dv);

        cdmach[i] = new CdMachPt(cd, mach);
    } // iterate over vtdata segments

    cdmach.sort(function (a,b) {return a.mach - b.mach});

    this.calculated = cdmach;

    // interpolate
    this.interpolated = [];

    var order = 2; // poly regression order

    var il, ih; // low and high indexes for points used for interpolation
    var bl, bh; // meibouring points in RMT table -- lower and higher
    var segment = [];
    var poly;
    var pt;

    var RMT = [];
    // fill in the reference table, rejecting points that are outside of the analysed data range
    for (let i = 0; i < RMT_FULL.length; i = i + 1) {
        if ( (RMT_FULL[i] > cdmach[0].mach) && (RMT_FULL[i] < cdmach[cdmach.length -1].mach) ) {
            RMT.push(RMT_FULL[i]);
        }
    }

    // calculate cd/mach values for reference Mach table, by interpolating the nearest data points
    for (let i = 0; i < RMT.length; i = i + 1) {
        // get low index
        il = 0;
        if (i == 0) { bl = 0; } else { bl = i - 1; } 
        while ( (cdmach[il].mach < RMT[bl]) && (il < cdmach.length - 1) ) { il = il + 1; }
        // get high index
        ih = cdmach.length - 1;
        if (i == (RMT.length - 1)) { bh = i; } else { bh = i + 1; }
        while ( (cdmach[ih].mach > RMT[bh]) && (ih > 0) ){ ih = ih - 1; }
        // check if enough points to interpolate, add if required
        while ((ih - il) < order) {
            if (il == 0) { ih = ih + 1; }
            else if (ih == cdmach.length - 1) { il = il - 1; }
            else if (Math.abs(cdmach[il-1].mach - RMT[i]) < Math.abs(cdmach[ih+1].mach - RMT[i])) { il = il - 1; }
            else { ih = ih + 1; }
        }
        // collect points for interpolation
        segment = [];
        for (let j = il; j <= ih; j = j + 1) {
            segment.push([cdmach[j].mach, cdmach[j].cd]);
        }
        poly = new RegressionPoly(segment, order);
        pt = new CdMachPt(poly.getPolyValue(RMT[i]), RMT[i]);
        this.interpolated.push(pt);
    }
} // function CdMachTable(veltab) constructor