/*
Copyright 2023 Alexandre Trofimov

This file is part of getcdmach external ballistics software.

getcdmach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

getcdmach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with getcdmach.  If not, see <http://www.gnu.org/licenses/>.
*/

// service routines for polynomial regression

function RegressionPoly (pts, order) {
    // calculates weighted least squares polynomial regression of order "order"
    // for all data points

    var m = [], // coefficients matrix
        r = []; // results vector

    var xpow = 0; // service
    var b = []; // w*(x^i) coefficients to be used to fill the M matrix
    b.length = 2 * order + 1;
    b.fill(0);
    var t = []; // product result vector
    t.length = order + 1;
    t.fill(0);

    var x, y, w;
    var i, j, k;

    // calculate B and T values
    var l = pts.length;
    for (i = 1; i < l; i = i + 1) {
        x = pts[i][0];
        y = pts[i][1];
        xpow = 1;

        for (k = 0; k <= order; k = k + 1) {
            b[k] = b[k] + xpow;
            t[order - k] = t[order - k] + xpow * y;
            xpow = xpow * x;
        }
        for (k = order + 1; k <= 2 * order; k = k + 1) {
            b[k] = b[k] + xpow;
            xpow = xpow * x;
        }
    }

    // fill in the m matrix
    for (i = 0; i <= order; i = i + 1) {
        m[i] = [];
        for (j = 0; j <= order; j = j + 1) {
            m[i][j] = b[2 * order - i - j];
        }
    }

    // solve for R in MR=T

    const NONZERO_THRESHOLD = 1e-30; // sanity check for non-degenerate matrix

    // LUP decomposition

    var maxVal, maxIdx, absVal, tmprow = [];
    var p = [];
    var n = t.length; // matrix and vector size for the problem at hand (= poly order + 1)
    for (i = 0; i < n; i = i + 1) { p[i] = i; }

    for (i = 0; i < n; i = i + 1) {
        maxVal = 0.0;
        maxIdx = i;

        // find the row with max first coeff
        for (j = i; j < n; j = j + 1) {
            absVal = Math.abs(m[j][i]);
            if (absVal > maxVal) {
                maxVal = absVal;
                maxIdx = j;
            }
        }

        if (maxVal < NONZERO_THRESHOLD) {
            // throw ("cannot fit poly regression, matrix is degenerate. bombing out.");
        }

        if (maxIdx != i) {
            // permute p
            j = p[i];
            p[i] = p[maxIdx];
            p[maxIdx] = j;

            // permute rows of m
            tmprow = m[i];
            m[i] = m[maxIdx];
            m[maxIdx] = tmprow;
        }

        for (j = i + 1; j < n; j = j + 1) {
            m[j][i] = m[j][i] / m[i][i];

            for (k = i + 1; k < n; k = k + 1)
                m[j][k] = m[j][k] - m[j][i] * m[i][k];
        }
    } // LUP decomposition loop

    // solve for r
    for (i = 0; i < n; i = i + 1) {
        r[i] = t[p[i]];
        for (j = 0; j < i; j = j + 1) {
            r[i] = r[i] - m[i][j] * r[j];
        }
    }

    for (i = n - 1; i >= 0; i = i - 1) {
        for (j = i + 1; j < n; j = j + 1) {
            r[i] = r[i] - m[i][j] * r[j];
        }
        r[i] = r[i] / m[i][i];
    }

    // regression done

    this.poly = r;

    this.getPolyValue = function (x) {
        // returns the f value corresponding to x argument, as per the poly model
        var order = this.poly.length - 1;
        var tpow = 1;
        var f = 0;
        for (let i = order; i >= 0; i = i - 1) {
            f = f + tpow * this.poly[i];
            tpow = tpow * x;
        }
        return (f);
    };
}; // regressionPoly(order,weight) method